@@ -1,25 +1,32 @@
 /*
  * tclxml.c --
  *
- *  Generic interface to XML parsers.
+ *  Entry point for XML parsers, DOM and XSLT.
  *
+ * Copyright (c) 2005-2007 Steve Ball, explain
+ * http://www.explain.com.au/
  * Copyright (c) 1998-2004 Steve Ball, Zveno Pty Ltd
  *
  * See the file "LICENSE" for information on usage and
  * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * $Id: tclxml.c,v 1.29 2004/02/26 05:12:21 balls Exp $
+ * $Id: tclxml.c,v 1.2 2009/03/03 22:55:36 joye Exp $
  *
  */
 
 #include <tclxml/tclxml.h>
+#include <tclxml-libxml2/tclxml-libxml2.h>
+#include <tcldom-libxml2/tcldom-libxml2.h>
+#include <tclxslt/tclxslt.h>
 #include <string.h>
 
 #define TCL_DOES_STUBS \
     (TCL_MAJOR_VERSION > 8 || TCL_MAJOR_VERSION == 8 && (TCL_MINOR_VERSION > 1 || \
     (TCL_MINOR_VERSION == 1 && TCL_RELEASE_LEVEL == TCL_FINAL_RELEASE)))
+#ifdef USE_TCLXML_STUBS
 #ifndef TCLXML_DOES_STUBS
 # define TCLXML_DOES_STUBS TCL_DOES_STUBS
+#endif /* USE_TCLXML_STUBS */
 #endif /* TCL_DOES_STUBS */
 
 /*
@@ -27,6 +34,8 @@
  */
 
 typedef struct ThreadSpecificData {
+  int initialized;
+
   TclXML_ParserClassInfo *defaultParser;    /* Current default parser */
   Tcl_HashTable *registeredParsers;         /* All known parser classes */
 
@@ -42,8 +51,14 @@ typedef struct ThreadSpecificData {
 
   int uniqueCounter;
 
+  /*
+   * Callback for external entity resolution
+   */
+
+  Tcl_Obj *externalentitycmd;
+  Tcl_Interp *interp;
 } ThreadSpecificData;
-/* static Tcl_ThreadDataKey dataKey; */
+static Tcl_ThreadDataKey dataKey;
 
 /* This string is a backup.  Value should be defined in xml package. */
 static char whitespace[] = " \t\r\n";
@@ -52,6 +67,14 @@ static char whitespace[] = " \t\r\n";
  * Configuration option tables
  */
 
+static CONST84 char *globalConfigureSwitches[] = {
+  "-externalentitycommand",
+  (char *) NULL
+};
+enum globalConfigureSwitches {
+  TCLXML_GLOBAL_EXTERNALENTITYCOMMAND
+};
+
 static CONST84 char *instanceConfigureSwitches[] = {
     "-final",
     "-validate",
@@ -113,7 +136,6 @@ enum instanceConfigureSwitches {
  * Prototypes for procedures defined later in this file:
  */
 
-static void DeletePkgData _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp));
 static void TclXMLInstanceDeleteCmd _ANSI_ARGS_((ClientData clientData));
 static int  TclXMLDestroyParserInstance _ANSI_ARGS_((TclXML_Info *xmlinfo));
 static int  TclXMLInstanceCmd _ANSI_ARGS_((ClientData dummy,
@@ -173,14 +195,16 @@ Tcl_GetString (obj)
  *
  * Tclxml_Init --
  *
- *  Initialisation routine for loadable module
+ *  Initialisation routine for loadable module.
+ *  Also calls the initialisation routines for TclDOM and TclXSLT,
+ *  as these were originally separate modules.
  *
  * Results:
  *  None.
  *
  * Side effects:
  *  Creates commands in the interpreter,
- *  loads xml package.
+ *  loads xml, dom and xslt packages.
  *
  *----------------------------------------------------------------------------
  */
@@ -189,9 +213,6 @@ int
 Tclxml_Init (interp)
       Tcl_Interp *interp; /* Interpreter to initialise. */
 {
-  /*
-  ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey);
-  */
   ThreadSpecificData *tsdPtr;
 
 #ifdef USE_TCL_STUBS
@@ -200,11 +221,12 @@ Tclxml_Init (interp)
   }
 #endif
 
-  tsdPtr = (ThreadSpecificData *) Tcl_Alloc(sizeof(ThreadSpecificData));
-  Tcl_SetAssocData(interp, "::xml::c", DeletePkgData, (ClientData) tsdPtr);
+  tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
+  tsdPtr->initialized = 1;
   tsdPtr->defaultParser = NULL;
   tsdPtr->uniqueCounter = 0;
 
+  /*
   tsdPtr->whitespaceRE = Tcl_GetVar2Ex(interp, "::xml::Wsp", NULL, TCL_GLOBAL_ONLY);
   if (tsdPtr->whitespaceRE == NULL) {
     tsdPtr->whitespaceRE = Tcl_SetVar2Ex(interp, "::xml::Wsp", NULL, Tcl_NewStringObj(whitespace, -1), TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG);
@@ -213,13 +235,31 @@ Tclxml_Init (interp)
     }
   }
   Tcl_IncrRefCount(tsdPtr->whitespaceRE);
+  */
 
   tsdPtr->registeredParsers = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
   Tcl_InitHashTable(tsdPtr->registeredParsers, TCL_STRING_KEYS);
 
-  Tcl_CreateObjCommand(interp, "xml::configure", TclXMLConfigureCmd, (ClientData) tsdPtr, NULL);
-  Tcl_CreateObjCommand(interp, "xml::parser", TclXMLCreateParserCmd, (ClientData) tsdPtr, NULL);
-  Tcl_CreateObjCommand(interp, "xml::parserclass", TclXMLParserClassCmd, (ClientData) tsdPtr, NULL);
+  tsdPtr->externalentitycmd = NULL;
+  tsdPtr->interp = interp;
+
+  Tcl_CreateObjCommand(interp, "xml::configure", TclXMLConfigureCmd, NULL, NULL);
+  Tcl_CreateObjCommand(interp, "xml::parser", TclXMLCreateParserCmd, NULL, NULL);
+  Tcl_CreateObjCommand(interp, "xml::parserclass", TclXMLParserClassCmd, NULL, NULL);
+
+  if (Tclxml_libxml2_Init(interp) != TCL_OK) {
+    return TCL_ERROR;
+  }
+
+  /*
+  if (Tcldom_libxml2_Init(interp) != TCL_OK) {
+    return TCL_ERROR;
+  }
+
+  if (Tclxslt_libxslt_Init(interp) != TCL_OK) {
+    return TCL_ERROR;
+  }
+  */
 
   #if TCLXML_DOES_STUBS
     {
@@ -262,36 +302,6 @@ Tclxml_SafeInit (interp)
     return Tclxml_Init(interp);
 }
 
-/*
- *----------------------------------------------------------------------------
- *
- * DeletePkgData --
- *
- *  Cleanup package memory usage.
- *
- * Results:
- *  None.
- *
- * Side effects:
- *  Memory deallocated.
- *
- *----------------------------------------------------------------------------
- */
-
-static void
-DeletePkgData (clientData, interp)
-     ClientData clientData;
-     Tcl_Interp *interp;
-{
-  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) clientData;
-
-  Tcl_DecrRefCount(tsdPtr->whitespaceRE);
-  Tcl_DeleteHashTable(tsdPtr->registeredParsers);
-  Tcl_Free((char *) tsdPtr->registeredParsers);
-
-  Tcl_Free((char *) tsdPtr);
-}
-
 /*
  *----------------------------------------------------------------------------
  *
@@ -315,7 +325,36 @@ TclXMLConfigureCmd(clientData, interp, objc, objv)
      int objc;
      Tcl_Obj *CONST objv[];
 {
-  /* Not yet implemented */
+  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
+  int index;
+
+  if (objc < 3) {
+    Tcl_SetResult(interp, "must specify option", NULL);
+    return TCL_ERROR;
+  } else if (objc == 3) {
+    /* TODO: retrieve option's value */
+    return TCL_OK;
+  } else if (objc % 2 == 1) {
+    Tcl_SetResult(interp, "value for option missing", NULL);
+    return TCL_ERROR;
+  }
+
+  for (objc -= 2, objv += 2; objc; objc -= 2, objv += 2) {
+    if (Tcl_GetIndexFromObj(interp, objv[0], globalConfigureSwitches,
+			    "switch", 0, &index) != TCL_OK) {
+      return TCL_ERROR;
+    }
+    switch ((enum globalConfigureSwitches) index) {
+    case TCLXML_GLOBAL_EXTERNALENTITYCOMMAND:
+      tsdPtr->externalentitycmd = objv[1];
+      Tcl_IncrRefCount(tsdPtr->externalentitycmd);
+      break;
+
+    default:
+      return TCL_ERROR;
+    }
+  }
+
   return TCL_OK;
 }
 
@@ -342,8 +381,7 @@ TclXMLParserClassCmd(clientData, interp, objc, objv)
      int objc;
      Tcl_Obj *CONST objv[];
 {
-  /* ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey); */
-  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) clientData;
+  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
   TclXML_ParserClassInfo *classinfo;
   int method, index;
   Tcl_Obj *listPtr;
@@ -541,13 +579,10 @@ TclXML_RegisterXMLParser(interp, classinfo)
      Tcl_Interp *interp;
      TclXML_ParserClassInfo *classinfo;
 {
-  /* ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey); */
-  ThreadSpecificData *tsdPtr;
+  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
   int new;
   Tcl_HashEntry *entryPtr;
 
-  tsdPtr = Tcl_GetAssocData(interp, "::xml::c", NULL);
-
   entryPtr = Tcl_CreateHashEntry(tsdPtr->registeredParsers, Tcl_GetStringFromObj(classinfo->name, NULL), &new);
   if (!new) {
     Tcl_Obj *ptr = Tcl_NewStringObj("parser class \"", -1);
@@ -593,8 +628,7 @@ TclXMLCreateParserCmd(clientData, interp, objc, objv)
      int objc;
      Tcl_Obj *CONST objv[];
 {
-  /* ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey); */
-  ThreadSpecificData *tsdPtr = Tcl_GetAssocData(interp, "::xml::c", NULL);
+  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
   TclXML_Info *xmlinfo;
   int found, i, index, poption;
 
@@ -606,6 +640,10 @@ TclXMLCreateParserCmd(clientData, interp, objc, objv)
     TCLXML_PARSER
   };
 
+  if (tsdPtr == NULL) {
+    Tcl_SetObjResult(interp, Tcl_NewStringObj("TclXML package improperly initialised", -1));
+    return TCL_ERROR;
+  }
   if (!tsdPtr->defaultParser) {
     Tcl_SetResult(interp, "no parsers available", NULL);
     return TCL_ERROR;
@@ -853,13 +891,12 @@ static Tcl_Obj *
 FindUniqueCmdName(interp)
      Tcl_Interp *interp;
 {
-  /* ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey); */
-  ThreadSpecificData *tsdPtr = Tcl_GetAssocData(interp, "::xml::c", NULL);
+  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
   Tcl_Obj *name;
   Tcl_CmdInfo cmdinfo;
   char s[20];
 
-  name = Tcl_NewStringObj("", 0);
+  name = Tcl_NewObj();
   Tcl_IncrRefCount(name);
 
   do {
@@ -1043,9 +1080,6 @@ TclXMLCreateEntityParser(interp, external, parent)
      TclXML_Info *external;
      TclXML_Info *parent;
 {
-  /*
-  TclXML_ParserClassInfo *parentClassInfo = (TclXML_ParserClassInfo *) parent->parserClass;
-  */
   TclXML_ParserClassInfo *extClassInfo;
 
   external->parserClass = parent->parserClass;
@@ -1273,7 +1307,6 @@ static void
 TclXMLFreeParser(xmlinfo)
      TclXML_Info *xmlinfo;
 {
-
   if (TclXMLDestroyParserInstance(xmlinfo) == TCL_OK) {
 	if (xmlinfo->encoding) {
 	  Tcl_DecrRefCount(xmlinfo->encoding);
@@ -1414,7 +1447,7 @@ TclXMLInstanceCmd (clientData, interp, objc, objv)
 	  if (strlen(encoding) == 0 || strcmp(encoding, "utf-8") == 0) {
 		data = Tcl_GetStringFromObj(objv[2], &len);
 	  } else {
-		data = Tcl_GetByteArrayFromObj(objv[2], &len);
+	    data = (char *) Tcl_GetByteArrayFromObj(objv[2], &len);
 	  }
 
       result = TclXMLParse(interp, xmlinfo, data, len);
@@ -1658,7 +1691,7 @@ TclXMLInstanceConfigure (interp, xmlinfo, objc, objv)
         if (Tcl_GetBooleanFromObj(interp, objv[1], &bool) != TCL_OK) {
           return TCL_ERROR;
         }
-        xmlinfo->validate = bool;
+        xmlinfo->expandinternalentities = bool;
         break;
 
       case TCLXML_PARAMENTITYPARSING:
@@ -2440,8 +2473,7 @@ static void
 TclXMLDispatchPCDATA(xmlinfo)
      TclXML_Info *xmlinfo;
 {
-  /* ThreadSpecificData *tsdPtr = TCL_TSD_INIT(&dataKey); */
-  ThreadSpecificData *tsdPtr = Tcl_GetAssocData(xmlinfo->interp, "::xml::c", NULL);
+  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
   int result = TCL_OK;
 
   if (xmlinfo->cdata == NULL || 
@@ -2847,9 +2879,11 @@ TclXML_UnknownEncodingHandler(encodingHandlerData, name, info)
  * TclXML_ExternalEntityRefHandler --
  *
  *  Called by parser instance for processing external entity references.
+ *  May also be called outside the context of a parser for XInclude
+ *  or XSLT import/include.
  *
  * Results:
- *  None.
+ *  Returns success code.
  *
  * Side Effects:
  *  Callback script is invoked.
@@ -2860,68 +2894,125 @@ TclXML_UnknownEncodingHandler(encodingHandlerData, name, info)
 int
 TclXML_ExternalEntityRefHandler(userData, openEntityNames, base,
     systemId, publicId)
-     ClientData userData;
+     ClientData userData;	/* NULL if not in parser context, current interp gets result */
      Tcl_Obj *openEntityNames;
      Tcl_Obj *base;
      Tcl_Obj *systemId;
      Tcl_Obj *publicId;
 {
+  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
   TclXML_Info *xmlinfo = (TclXML_Info *) userData;
   int result = TCL_OK;
   Tcl_Obj *oldContext;
 
-  TclXMLDispatchPCDATA(xmlinfo);
+  if (xmlinfo != NULL) {
+    TclXMLDispatchPCDATA(xmlinfo);
 
-  if ((xmlinfo->entitycommand == NULL && xmlinfo->entity == NULL) || 
-      xmlinfo->status != TCL_OK) {
-      return 0;
-  }
-  oldContext = xmlinfo->context;
-  xmlinfo->context = openEntityNames;
-
-  if (xmlinfo->entity) {
-    result = (xmlinfo->entity)(xmlinfo->interp, xmlinfo->entitydata, xmlinfo->name, base, systemId, publicId);
-  } else if (xmlinfo->entitycommand) {
-    Tcl_Obj *cmdPtr;
-
-    /*
-     * Take a copy of the callback script so that arguments may be appended.
-     */
-
-    cmdPtr = Tcl_DuplicateObj(xmlinfo->entitycommand);
-    Tcl_IncrRefCount(cmdPtr);
-    Tcl_Preserve((ClientData) xmlinfo->interp);
-
-    Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, xmlinfo->name);
-
-    if (base) {
-      Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, base);
-    } else {
-      Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, Tcl_NewObj());
+    if (xmlinfo->entitycommand == NULL && xmlinfo->entity == NULL) {
+      if (Tcl_IsSafe(xmlinfo->interp)) {
+	return TCL_BREAK;
+      } else {
+	return TCL_CONTINUE;
+      }
     }
+    if (xmlinfo->status != TCL_OK) {
+      return xmlinfo->status;
+    }
+    oldContext = xmlinfo->context;
+    xmlinfo->context = openEntityNames;
 
-    Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, systemId);
+    if (xmlinfo->entity) {
+      result = (xmlinfo->entity)(xmlinfo->interp, xmlinfo->entitydata, xmlinfo->name, base, systemId, publicId);
+    } else if (xmlinfo->entitycommand) {
+      Tcl_Obj *cmdPtr;
 
-    if (publicId) {
-      Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, publicId);
-    } else {
-      Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, Tcl_NewObj());
+      /*
+       * Take a copy of the callback script so that arguments may be appended.
+       */
+
+      cmdPtr = Tcl_DuplicateObj(xmlinfo->entitycommand);
+      Tcl_IncrRefCount(cmdPtr);
+      Tcl_Preserve((ClientData) xmlinfo->interp);
+
+      Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, xmlinfo->name);
+
+      if (base) {
+	Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, base);
+      } else {
+	Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, Tcl_NewObj());
+      }
+
+      Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, systemId);
+
+      if (publicId) {
+	Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, publicId);
+      } else {
+	Tcl_ListObjAppendElement(xmlinfo->interp, cmdPtr, Tcl_NewObj());
+      }
+
+      /*
+       * It would be desirable to be able to terminate parsing
+       * if the return result is TCL_ERROR or TCL_BREAK.
+       */
+      result = Tcl_GlobalEvalObj(xmlinfo->interp, cmdPtr);
+
+      Tcl_DecrRefCount(cmdPtr);
+      Tcl_Release((ClientData) xmlinfo->interp);
     }
 
     /*
-     * It would be desirable to be able to terminate parsing
-     * if the return result is TCL_ERROR or TCL_BREAK.
+     * Return results have a different meaning for external entities,
+     * so don't retain the result for later use.
+     * TclXMLHandlerResult(xmlinfo, result);
      */
-    result = Tcl_GlobalEvalObj(xmlinfo->interp, cmdPtr);
+    xmlinfo->context = oldContext;
 
-    Tcl_DecrRefCount(cmdPtr);
-    Tcl_Release((ClientData) xmlinfo->interp);
+  } else {
+    /*
+     * No parser context
+     */
+
+    if (tsdPtr->externalentitycmd) {
+      Tcl_Obj *cmdPtr;
+
+      /*
+       * Take a copy of the callback script so that arguments may be appended.
+       */
+
+      cmdPtr = Tcl_DuplicateObj(tsdPtr->externalentitycmd);
+      Tcl_IncrRefCount(cmdPtr);
+      Tcl_Preserve((ClientData) tsdPtr->interp);
+
+      if (base) {
+	Tcl_ListObjAppendElement(tsdPtr->interp, cmdPtr, base);
+      } else {
+	Tcl_ListObjAppendElement(tsdPtr->interp, cmdPtr, Tcl_NewObj());
+      }
+
+      Tcl_ListObjAppendElement(tsdPtr->interp, cmdPtr, systemId);
+
+      if (publicId) {
+	Tcl_ListObjAppendElement(tsdPtr->interp, cmdPtr, publicId);
+      } else {
+	Tcl_ListObjAppendElement(tsdPtr->interp, cmdPtr, Tcl_NewObj());
+      }
+
+      /*
+       * It would be desirable to be able to terminate parsing
+       * if the return result is TCL_ERROR or TCL_BREAK.
+       */
+      result = Tcl_GlobalEvalObj(tsdPtr->interp, cmdPtr);
+
+      Tcl_DecrRefCount(cmdPtr);
+      Tcl_Release((ClientData) tsdPtr->interp);
+    } else if (Tcl_IsSafe(tsdPtr->interp)) {
+      return TCL_BREAK;
+    } else {
+      return TCL_CONTINUE;
+    }
   }
 
-  TclXMLHandlerResult(xmlinfo, result);
-  xmlinfo->context = oldContext;
-  
-  return 1;
+  return result;
 }
 
 /*
